#!/usr/bin/perl

####################################################
#
# DISSPAM.PL
#
# VERSION 0.10
#
# Just one more script that tries to combat spam
#
# Written by Mina Naguib [webmaster@topfx.com]
#
####################################################

use strict;
use Net::POP3;
use Net::DNS;
use vars qw($CONFFILENAME %RBL %MAILBOX $SENDMAIL $DATAFILE $PIDFILE %HEADERHASHES);

$CONFFILENAME = shift || &usage();
(-f $CONFFILENAME) || die "Error: $CONFFILENAME is not a file\n";
&parseconffile();

&loadheaderhashes();

&checkemail();

&saveheaderhashes();

&deletepidfile();

#
# This sub is the main loop. It loops through all the mailboxes defined in hash %MAILBOX
# , checks them for spam, and does whatever's appropriate if spam's found (backup to
# file / delete message / send bounceback etc...)
#
sub checkemail() {
	my $host;
	my $username;
	my $password;
	my $email;
	my $pop;
	my $temp;
	my $messages;
	my $message;
	my $headers;
	my $reason;
	my $mbox;
	local (*FH);
	foreach $mbox (sort keys %MAILBOX) {
		print "Processing [$mbox] ...\n";
		print "\n";
		$username = $MAILBOX{$mbox}{username};
		if (!$username) {
			warn "Error: Configuration [$mbox] is missing username. Skipping.\n";
			next;
			}
		$password = $MAILBOX{$mbox}{password};
		if (!$password) {
			warn "Error: Configuration [$mbox] is missing password. Skipping.\n";
			next;
			}
		$host = $MAILBOX{$mbox}{host};
		if (!$host) {
			warn "Error: Configuration [$mbox] is missing host. Skipping.\n";
			next;
			}
		$email = $MAILBOX{$mbox}{email};
		if (!$email) {
			warn "Error: Configuration [$mbox] is missing email. Skipping.\n";
			next;
			}
		$pop = Net::POP3->new($host, Timeout => 30);
		if (!$pop) {
			warn "Error: Could not connect to $host. Skipping.\n";
			next;
			}
		$temp = $pop->login($username, $password);
		if (!$temp) {
			warn "Error: Could not login to $host. Check username and password in configuration file. Skipping.\n";
			$pop->quit();
			next;
			}
		$messages = $pop->list();
		if (!$messages) {
			warn "Error: Could not retrieve message list\n";
			$pop->quit();
			next;
			}
		foreach (keys %{$messages}) {
			$headers = $pop->top($_);
			if (ref($headers) ne "ARRAY") {
				next;
				}
			$headers = join("", @$headers);
			if ($reason = &checkheaders($headers, $mbox)) {
				print "\tTHAT IS SPAM\n";
				if ($MAILBOX{$mbox}{deletespam}) {
					if ($MAILBOX{$mbox}{backupfile}) {
						print "\tRETRIEVING ENTIRE MESSAGE\n";
						$message = $pop->get($_);
						$message = join("", @$message);
						print "\tSAVING MESSAGE TO BACKUP FILE\n";
						open (FH, ">>$MAILBOX{$mbox}{backupfile}") || die "Error: Could not open() $MAILBOX{$mbox}{backupfile} for appending: $!\n";
						print FH "From inet\n$message\n";
						close (FH);
						}
					if ($MAILBOX{$mbox}{sendbounceback}) {
						&sendbounceback($email, $headers, $reason);
						}
					$pop->delete($_);
					print "\tDELETED FROM SERVER\n";
					}
				else {
					print "\tNOT DELETED FROM SERVER\n";
					}
				}
			else {
				print "\tTHAT IS NOT SPAM\n";
				if ($MAILBOX{$mbox}{fetchcleanmessagesto}) {
					if ($MAILBOX{$mbox}{fetchcleanmessagesto} eq $MAILBOX{$mbox}{email}) {
						print "\tTHE FETCH-TO ADDRESS IS THE SAME AS THE MAILBOX ADDRESS\n";
						print "\tFETCH ABORTED TO AVOID LOOPBACK\n";
						}
					elsif (!$SENDMAIL) {
						print "\tSENDMAIL NOT CONFIGURED. NOT FETCHING CLEAN MESSAGE\n";
						}
					elsif (!&validemail($MAILBOX{$mbox}{fetchcleanmessagesto})) {
						print "\tTHE FETCH-TO ADDRESS [$MAILBOX{$mbox}{fetchcleanmessagesto}] IS INVALID. NOT FETCHING CLEAN MESSAGE\n";
						}
					else {
						print "\tRETRIEVING ENTIRE MESSAGE\n";
						$message = $pop->get($_);
						$message = join("", @$message);
						$message =~ s/^To:.*$/To: $MAILBOX{$mbox}{fetchcleanmessagesto}/mi;
						print "\tRE-SENDING MESSAGE TO $MAILBOX{$mbox}{fetchcleanmessagesto}\n";
						open (FH, "| $SENDMAIL -ti") || die "Error: Couldn't start SENDMAIL [$SENDMAIL]: $!\n";
						print FH $message;
						close (FH);
						print "\tDELETING MESSAGE FROM SERVER\n";
						$pop->delete($_);
						}
					}
				}
			print "\n";
			}
		$temp = $pop->quit();
		print "\n";
		}
	}

#
# This sub takes a scalar email headers and the mailbox hash key
# parses the header and prints diagnostics, then returns true (an error message)
# If the message is spam, or 0 if it's not.
#
sub checkheaders() {
	my $headers = shift;
	my $mbox = shift;
	my $hash;
	my @header;
	my $from;
	my $subject;
	my %ip;
	my $temp;
	foreach(split(/\n+/, $headers)) {
		if (/^\s/) {
			$header[$#header] .= $_;
			}
		else {
			push (@header, $_);
			}
		}
	foreach (@header) {
		if (/^Received:/ && /\[(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\]/) {
			$ip{$1} = 1;
			}
		elsif (/^From:\s*(.+?)\s*$/) {
			$from = $1;
			print "\tFrom: $from\n";
			foreach (@{$MAILBOX{$mbox}{exclude_from}}) {
				if ($from =~ /$_/i) {
					print "\tEXCLUDING $_\n";
					return 0;
					}
				}
			foreach (@{$MAILBOX{$mbox}{ban_from}}) {
				if ($from =~ /$_/i) {
					print "\tBAN FROM MATCHED $_\n";
					return "Sender is banned from emailing this recipient";
					}
				}
			}
		elsif (/^Subject:\s*(.+?)\s*$/) {
			$subject = $1;
			print "\tSubject: $subject\n";
			foreach (@{$MAILBOX{$mbox}{ban_subject}}) {
				if ($subject =~ /$_/i) {
					print "\tBAN SUBJECT MATCHED $_\n";
					return "Message genre/subject is banned from being received by this recipient";
					}
				}
			}
		}
	if ($MAILBOX{$mbox}{checkonlyonce}) {
		$hash = &gethash($headers);
		if ($HEADERHASHES{$hash}) {
			$HEADERHASHES{$hash} = 2;
			print "\tCHECKED BEFORE.  SKIPPING.\n";
			return 0;
			}
		else {
			$HEADERHASHES{$hash} = 2;
			}
		}
	foreach (keys %ip) {
		if ($temp = &banned($_, $mbox)) {
			return "An internet e-mail server this message passed through [$_] is banned by RBL blacklist [$temp]";
			}
		}
	return 0;
	}

#
# This sub takes an IP address and a mailbox hash key
# and returns true (1 or rbl name) if the address is banned in an RBL
# or returns 0 if it's not banned
#
sub banned() {
	my $ip = shift || return "Empty mailserver IP";
	my $mbox = shift;
	my $dns = new Net::DNS::Resolver;
	my $query;
	my $entry;
	my $rblquery;
	my $reversedip = join(".", reverse split(/\./, $ip));
	foreach (@{$MAILBOX{$mbox}{exclude_ip}}) {
		if ($ip eq $_) {
			print "\tEXCLUDING $ip\n";
			return 0;
			}
		}
	foreach (keys %RBL) {
		$rblquery = "$reversedip.$_";
		if ($query = $dns->search($rblquery)) {
			foreach $entry ($query->answer) {
				if ($entry->type eq "A" && $entry->address =~ /^$RBL{$_}$/) {
					print "\tMATCHED $ip AT $_ (", $entry->address, ")\n";
					return $_;
					}
				}
			}
		}
	return 0;
	}

sub usage() {
	print <<"EOM";

Usage:

\t$0 file.conf

Where file.conf is the filename to read the configuration from.

See sample.conf for a sample of a configuration file, or readme.txt for documentation on configuration file syntax.

EOM
	exit(0);
	}

#
# This sub parses the config file, populating %RBL, %MAILBOX, $SENDMAIL, $DATAFILE, $PIDFILE
# Also does sanity checking and dies if it sees errors.
#
sub parseconffile() {
	local (*FH);
	my $sectionname;
	my $key;
	my $value;
	my %global;

	%MAILBOX = ();

	open (FH, $CONFFILENAME) || die "Error: Could not open() $CONFFILENAME : $!\n";
	while (<FH>) {
		chomp;
		if (/^#/ || !$_) {
			next;
			}
		if (/^\s*\[\s*(.+?)\s*\]\s*$/) {
			if (exists $MAILBOX{lc($1)}) {
				die "Error: Configuration file contains more than 1 section with name ", lc($1), "\n";
				}
			else {
				$sectionname = lc($1);
				}
			}
		elsif (/^\s*(.+?)\s*=\s*(.+?)\s*$/) {
			$key = lc($1);
			$value = $2;
			if ($key =~ /^(checkonlyonce|sendbounceback)$/) {
				#
				# We want these values to default to false if unspecified:
				#
				$value = ($value =~ /^(y|1)/i) ? 1 : 0;
				}
			elsif ($key =~ /^(deletespam)$/) {
				#
				# We want these values to default to true if unspecified:
				#
				$value = ($value =~ /^(n|0)/i) ? 0 : 1;
				}
			elsif ($key =~ /^(exclude_from|exclude_ip|ban_from|ban_subject)$/) {
				#
				# We want to convert these values to arrays (anonymous to be stored in hashkey)
				#
				$value = [split(/\s*,\s*/, $value)];
				}
			$MAILBOX{$sectionname}{$key} = $value;
			}
		}
	close (FH);
	#
	# We check that they've defined an rbl and global section otherwise perl'll crap out when we access it
	# as a hash
	#
	if (!exists $MAILBOX{rbl}) {
		die "Error: Your configuration file does not contain an [RBL] section\n";
		}
	elsif (!exists $MAILBOX{global}) {
		die "Error: Your configuration file does not contain a [GLOBAL] section\n";
		}
	#
	# We remove keys from %MAILBOX that are internal and not real mailbox definitions
	%RBL = %{$MAILBOX{rbl}};
	delete $MAILBOX{rbl};
	%global = %{$MAILBOX{global}};
	delete $MAILBOX{global};
	#
	# We extract non-inheritable values from %global
	#
	$SENDMAIL = $global{sendmail};
	delete $global{sendmail};
	$DATAFILE = $global{datafile};
	delete $global{datafile};
	$PIDFILE = $global{pidfile};
	delete $global{pidfile};
	#
	# Now we inherit the settings from %global if they're not defined per mailbox
	#
	foreach $key (keys %global) {
		foreach (keys %MAILBOX) {
			$MAILBOX{$_}{$key} = (exists $MAILBOX{$_}{$key}) ? $MAILBOX{$_}{$key} : $global{$key};
			}
		}
	if (!$PIDFILE) {
		die "Error: You must specify a PIDFILE directive in the [GLOBAL] section in the configuration file\n";
		}
	elsif (-f $PIDFILE) {
		die "Error: Another instance of DisSpam is already running.  Please wait till it finishes.  If no DisSpam is currently running please delete PIDFILE [$PIDFILE]\n";
		}
	elsif (!scalar keys %RBL) {
		die "Error: Configuration [RBL] section does not contain any RBL hostnames.\n";
		}
	elsif (!scalar keys %MAILBOX) {
		die "Error: Configuration file does not define any mailboxes\n";
		}
	elsif ($SENDMAIL && (!-f $SENDMAIL || !-x $SENDMAIL)) {
		die "Error: Configured SENDMAIL [$SENDMAIL] does not exist or is not executable by me\n";
		}
	elsif (!$DATAFILE) {
		die "Error: You must specify a DATAFILE directive in the [GLOBAL] section in the configuration file\n";
		}
	if (!open(FH, ">>$DATAFILE")) {
		die "Error: Could not open DATAFILE [$DATAFILE] for writing: $!\n";
		}
	else {
		close (FH);
		}
	if (!open(FH, ">$PIDFILE")) {
		die "Error: Could not open PIDFILE [$PIDFILE] for writing: $!\n";
		}
	else {
		print FH "$$\n";
		close (FH);
		}
	}

#
# This sub takes in an email address, message headers, and a reason why the message was not delivered
# It then sends back an appropriate email (bounceback) saying the mail was not delivered and why
#
sub sendbounceback() {
	my $email = shift;
	my $headers = shift;
	my $reason = shift;
	my $daemon;
	my $daemon2;
	my $from;
	my $replyto;
	local (*SMH);

	if (!$SENDMAIL) {
		die "Error: Could not send bounceback because no SENDMAIL configuration directive exists in the [global] section in configuration file\n";
		}

	($replyto) = $headers =~ /^Reply-to:\s*(.+)$/mi;
	($from) = $headers =~ /^From:\s*(.+)$/mi;
	$from = ($replyto) ? $replyto : $from;
	if ($from =~ /([a-z0-9._-]+\@([a-z0-9-]+\.)+[a-z]{2,5})/i) {
		$from = $1;
		}
	else {
		return 0;
		}

	if ($headers =~ /^X-Loop/mi) {
		return 0;
		}
	elsif ($headers =~ /^Return-path:\s*<>/mi) {
		return 0;
		}
	elsif ($headers =~ /^Precedence:\s*(bulk|list|junk)/mi) {
		return 0;
		}

	$daemon = $email;
	$daemon =~ s/^(.+)\@(.+)$/MAILER-DAEMON\@$2/;
	$daemon2 = $from;
	$daemon2 =~ s/^(.+)\@(.+)$/MAILER-DAEMON\@$2/;

	print "\tSENDING BOUNCEBACK TO [$from]\n";
	open (SMH, "| $SENDMAIL -tiN never") || die "Error: Could not launch SENDMAIL [$SENDMAIL]: $!\n";
	print SMH <<"EOM";
Return-path: <>
Errors-To: $daemon2
From: Mail Delivery Subsystem <$daemon>
Reply-to: $daemon2
To: $from
Subject: Returned mail: see transcript for details
X-Loop: $email
Auto-Submitted: auto-generated (failure)



This message was automatically generated by the
anti-SPAM software DisSpam.



A message you have sent to this address:

	* $email

Was not delivered due to the following reason:

	* $reason

More information about spam RBL blacklists in general:

	* http://www.ordb.org/faq/

This was diagnosed by the DisSpam utility monitoring
this email address.

To correct this problem, you may try to do the
following:

	* Study the reason mentioned above. Sometimes
	  it's self explanatory (you're banned, subject
	  of your message is banned etc...)
	* Re-send your message through another account
	  that does not use the above blacklisted
	  mailserver, sender name, subject...
	* Contact the owner of this mailbox via other
	  means and ask them to add you to their
	  exceptions list.
	* Contact the administrator of the above
	  blacklisted mailserver and ask them to
	  investigate this problem.  Including this
	  message would probably help them.


Please do not reply to this email.


Thank you.

DisSpam Anti-Spam Daemon
http://freshmeat.net/projects/disspam/


Complete headers from your message are included below :
--

$headers

EOM
	close (SMH);
	}

#
# This sub takes in a scalar data and returns a unique hash corresponding to it
# (this is NOT fancy whatsoever. MD5 is probably much better, but I wanted less dependencies)
#
sub gethash() {
	my $data = shift;
	my $hash = unpack("%16C*", $data);
	return $hash;
	}

#
# This sub loads the datafile into the hash %HEADERHASHES
#
sub loadheaderhashes() {
	local (*FH);
	open (FH, $DATAFILE) || die "Error: Could not open DATAFILE [$DATAFILE] for input: $!\n";
	while (<FH>) {
		chomp;
		$HEADERHASHES{$_} = 1;
		}
	close (FH);
	}

#
# This sub saves the hash %HEADERHASHES into the datafile to be read the next run
# It only saves keys with a value of 2 (they exist in the mailbox(es)).
# Keys with a value of 1 mean we read them from the datafile originally, but they no
# longer exist in the mailboxes, so no need to keep them laying around
#
sub saveheaderhashes() {
	local (*FH);
	open (FH, ">$DATAFILE") || die "Error: Could not open DATAFILE [$DATAFILE] for output: $!\n";
	foreach (keys %HEADERHASHES) {
		if ($HEADERHASHES{$_} > 1) {
			print FH "$_\n";
			}
		}
	close (FH);
	}
#
# This sub deletes the PIDFILE we created in the beginning
#
sub deletepidfile() {
	if (!-f $PIDFILE) {
		die "Error: Could not delete PIDFILE [$PIDFILE] because it does not exist\n";
		}
	elsif (!unlink($PIDFILE)) {
		die "Error: Could not delete PIDFILE [$PIDFILE]: $!\n";
		}
	}

#
# This sub takes an email address and returns if it's valid or not
#
sub validemail() {
	my $email = shift;
	if ($email =~ /^[a-z0-9._-]+\@([a-z0-9-]+\.)+[a-z]{2,4}$/i) {
		return 1;
		}
	else {
		return 0;
		}
	}
